# -*- coding: utf-8 -*-

# built ins
import asyncio
import json
import time
import os, sys

# installed
import orjson
import websockets
from dataclassy import dataclass, fields
from loguru import logger as log

# -----------------------------------------------------------------------------

this_folder = os.path.dirname(os.path.abspath(__file__))
root_folder = os.path.dirname(os.path.dirname(this_folder))
sys.path.append(root_folder + "/python")
sys.path.append(this_folder)

# -----------------------------------------------------------------------------

# user defined formula
from streaming_helper.configuration import config
from streaming_helper.messaging.telegram_bot import telegram_bot_sendtext
from streaming_helper.utilities import error_handling


def parse_dotenv(sub_account: str) -> dict:
    return config.main_dotenv(sub_account)


def get_timestamp():
    return int(time.time() * 1000)


@dataclass(unsafe_hash=True, slots=True)
class StreamingDataBinance:
    """
    https://www.binance.com/en/support/faq/detail/18c97e8ab67a4e1b824edd590cae9f16

    """

    sub_account_id: str
    client_id: str = fields
    client_secret: str = fields
    # Async Event Loop
    loop = asyncio.get_event_loop()
    ws_connection_notice_url: str = "wss://bstream.binance.com:9443/stream?"
    # Instance Variables
    websocket_client: websockets.WebSocketClientProtocol = None
    refresh_token: str = None
    refresh_token_expiry_time: int = None

    async def ws_manager(
        self,
        client_redis: object,
        exchange: str,
        redis_channels,
        queue_general: object,
    ) -> None:

        async with websockets.connect(
            self.ws_connection_notice_url,
            ping_interval=None,
            compression=None,
            close_timeout=60,
        ) as self.websocket_client:

            try:

                # timestamp = get_timestamp()

                # encoding_result = hashing(timestamp, self.client_id, self.client_secret)

                msg = {
                    "method": "SUBSCRIBE",
                    "params": ["abnormaltradingnotices"],
                    "id": 1,
                }

                while True:

                    ws_channel = ["abnormaltradingnotices"]

                    await self.ws_operation(
                        operation="SUBSCRIBE",
                        ws_channel=ws_channel,
                        source="ws",
                    )

                    while True:

                        # Receive WebSocket messages
                        message: bytes = await self.websocket_client.recv()
                        message: dict = orjson.loads(message)

                        if message:

                            data = message.get("data", None)

                            if data:

                                data.update({"exchange": exchange})
                                data.update({"account_id": self.sub_account_id})

                                # queing message to dispatcher
                                await queue_general.put(data)

            except Exception as error:

                await error_handling.parse_error_message_with_redis(
                    client_redis,
                    error,
                )

    async def ws_operation(
        self,
        operation: str,
        ws_channel: str,
        source: str = "ws",
    ) -> None:
        """ """
        sleep_time: int = 0.05

        await asyncio.sleep(sleep_time)

        id = 1

        msg: dict = {}

        if "ws" in source:
            extra_params: dict = dict(
                id=id,
                method=f"{operation}",
                params=ws_channel,
            )

            msg.update(extra_params)

            await self.websocket_client.send(json.dumps(msg))



# -*- coding: utf-8 -*-

"""
why aiohttp over httpx?
    - Our module is fully using asynchronous which is aiohttp spesialization
    - has more mature asyncio support than httpx
    - aiohttp is more suitable for applications that require high concurrency and low latency, such as web scraping or real-time data processing.

references:
    - https://github.com/encode/httpx/issues/3215#issuecomment-2157885121
    - https://github.com/encode/httpx/discussions/3100
    - https://brightdata.com/blog/web-data/requests-vs-httpx-vs-aiohttp


"""

# built ins
import asyncio

# installed
import aiohttp
from aiohttp.helpers import BasicAuth

# user defined formula
from streaming_helper.restful_api.deribit import end_point_params_template as end_point_deribit
from streaming_helper.restful_api.telegram import (
    end_point_params_template as telegram_end_point,
)


async def get_connected(
    connection_url: str,
    endpoint: str = None,
    client_id: str = None,
    client_secret: str = None,
    params: str = None,
) -> None:


    async with aiohttp.ClientSession() as session:

        if endpoint:
            
            from loguru import logger as log

            log.debug(f"connection_url {connection_url} endpoint {endpoint}")
            connection_endpoint = connection_url + endpoint
            
            log.debug(f"connection_endpoint {connection_endpoint} ")

        if client_id:

            if "telegram" in connection_url:

                response = await telegram_response(
                    session,
                    connection_url,
                    endpoint,
                    client_id,
                    client_secret,
                    params,
                )

            if "deribit" in connection_url:

                response: dict = await deribit_response(
                    session,
                    connection_endpoint,
                    endpoint,
                    client_id,
                    client_secret,
                    params,
                )

        else:

            async with session.get(connection_endpoint) as response:

                # RESToverHTTP Response Content
                response: dict = await response.json()

        return response


async def telegram_response(
    session: object,
    connection_url: str,
    endpoint: str = None,
    client_id: str = None,
    client_secret: str = None,
    params: str = None,
) -> None:
    """ """
    endpoint = telegram_end_point.message_end_point(client_id, client_secret, params)

    connection_endpoint = connection_url + endpoint

    async with session.get(connection_endpoint) as response:

        # RESToverHTTP Response Content
        return await response.json()


async def deribit_response(
    session: object,
    connection_endpoint,
    endpoint: str = None,
    client_id: str = None,
    client_secret: str = None,
    params: str = None,
) -> None:
    
    from loguru import logger as log

    payload: dict = end_point_deribit.get_json_payload(
        endpoint,
        params,
    )   
    
    log.warning(f"payload {payload} endpoint {endpoint} params {params} ")
    log.debug(f"client_id {client_id} client_secret {client_secret} connection_endpoint {connection_endpoint} ")

    async with session.post(
        connection_endpoint,
        auth=BasicAuth(client_id, client_secret),
        json=payload,
    ) as response:

        # RESToverHTTP Response Content
        return await response.json()





# -*- coding: utf-8 -*-

import asyncio

from streaming_helper.restful_api.deribit import end_point_params_template as end_point
from streaming_helper.utilities.pickling import read_data
from streaming_helper.utilities.system_tools import (
    provide_path_for_file,
)


def reading_from_pkl_data(
    end_point: str,
    currency: str,
    status: str = None,
) -> dict:
    """ """

    path: str = provide_path_for_file(end_point, currency, status)
    return read_data(path)


async def combining_ticker_data(instruments_name: str) -> list:
    """_summary_
    https://blog.apify.com/python-cache-complete-guide/]
    https://medium.com/@jodielovesmaths/memoization-in-python-using-cache-36b676cb21ef
    data caching
    https://medium.com/@ryan_forrester_/python-return-statement-complete-guide-138c80bcfdc7

    Args:
        instrument_ticker (_type_): _description_

    Returns:
        _type_: _description_
    """

    result = []
    for instrument_name in instruments_name:

        result_instrument = reading_from_pkl_data("ticker", instrument_name)

        if result_instrument:
            result_instrument = result_instrument[0]

        else:
            
            connection_url = end_point.basic_https()

            endpoint_tickers = end_point.get_tickers_end_point(instrument_name)

            result_instrument = await connector.get_connected(
                connection_url,
                endpoint_tickers,
            )
            log.debug(f"ticker {result_instrument}")
            
        result.append(result_instrument)

    return result


async def update_cached_ticker(
    instrument_name: str,
    ticker: list,
    data_orders: dict,
) -> None:
    """_summary_
    https://stackoverflow.com/questions/73064997/update-values-in-a-list-of-dictionaries

    Args:
        instrument_ticker (_type_): _description_

    Returns:
        _type_: _description_
    """

    instrument_ticker: list = [
        o for o in ticker if instrument_name in o["instrument_name"]
    ]

    if instrument_ticker:

        for item in data_orders:

            if (
                "stats" not in item
                and "instrument_name" not in item
                and "type" not in item
            ):
                [o for o in ticker if instrument_name in o["instrument_name"]][0][
                    item
                ] = data_orders[item]

            if "stats" in item:

                data_orders_stat = data_orders[item]

                for item in data_orders_stat:
                    [o for o in ticker if instrument_name in o["instrument_name"]][0][
                        "stats"
                    ][item] = data_orders_stat[item]

    else:
        from loguru import logger as log

        log.warning(f" {ticker}")
        log.debug(f" {instrument_name}")

        log.critical(f"instrument_ticker before {instrument_ticker}")
        # combining_order_data(currencies)
        log.debug(f"instrument_ticker after []-not ok {instrument_ticker}")


def update_cached_orders(
    orders_all: list,
    sub_account_data: dict,
    source: str = "ws",
):
    """_summary_

    source: ws/rest
    https://stackoverflow.com/questions/73064997/update-values-in-a-list-of-dictionaries

    Args:
        instrument_ticker (_type_): _description_

    Returns:
        _type_: _description_
    """

    if source == "ws":

        try:

            orders = sub_account_data["orders"]

            trades = sub_account_data["trades"]

            if orders:

                if trades:

                    for trade in trades:

                        order_id = trade["order_id"]

                        selected_order = [
                            o for o in orders_all if order_id in o["order_id"]
                        ]

                        if selected_order:

                            orders_all.remove(selected_order[0])

                if orders:

                    for order in orders:

                        order_state = order["order_state"]

                        if order_state == "cancelled" or order_state == "filled":

                            order_id = order["order_id"]

                            selected_order = [
                                o for o in orders_all if order_id in o["order_id"]
                            ]

                            if selected_order:

                                orders_all.remove(selected_order[0])

                        else:

                            orders_all.append(order)
        except:

            from loguru import logger as log

            log.debug(sub_account_data)
            try:
                order = sub_account_data[0]

            except:
                order = sub_account_data

            try:
                order_state = order["order_state"]
            except:
                order_state = order["state"]

            if order_state == "cancelled" or order_state == "filled":

                order_id = order["order_id"]

                selected_order = [o for o in orders_all if order_id in o["order_id"]]

                if selected_order:

                    orders_all.remove(selected_order[0])

            else:

                orders_all.append(order)

    if source == "rest":

        orders = sub_account_data

        trades = []


def positions_updating_cached(
    positions_cached: list,
    sub_account_data: list,
    source: str = "ws",
):
    """ """

    if source == "ws":
        positions = sub_account_data["positions"]

    if source == "rest":

        positions = sub_account_data

    if positions:

        for position in positions:

            position_instrument_name = position["instrument_name"]

            selected_position = [
                o
                for o in positions_cached
                if position_instrument_name in o["instrument_name"]
            ]

            if selected_position:

                positions_cached.remove(selected_position[0])

            positions_cached.append(position)



# -*- coding: utf-8 -*-

# built ins
import asyncio

# user defined formula
from streaming_helper.db_management import redis_client
from streaming_helper.messaging import telegram_bot as tlgrm
from streaming_helper.utilities import string_modification as str_mod, system_tools


async def caching_distributing_data(
    client_redis: object,
    redis_channels: list,
    queue_general: object,
) -> None:
    """ """

    try:

        # preparing redis connection
        pubsub: object = client_redis.pubsub()

        abnormal_trading_notices_channel: str = redis_channels[
            "abnormal_trading_notices"
        ]

        result: dict = str_mod.message_template()

        while True:

            message_params: str = await queue_general.get()

            async with client_redis.pipeline() as pipe:

                try:

                    message_channel: str = message_params.get("stream")

                    if message_channel:

                        data: dict = message_params["data"]

                        pub_message = dict(
                            data=data,
                        )

                        if "abnormaltradingnotices" in message_channel:

                            data: dict = message_params["data"]

                            pub_message = dict(
                                data=data,
                            )

                            await abnormal_trading_notices_in_message_channel(
                                pipe,
                                abnormal_trading_notices_channel,
                                pub_message,
                                result,
                            )

                        await pipe.execute()

                except Exception as error:

                    await system_tools.parse_error_message_with_redis(
                        client_redis,
                        error,
                    )

    except Exception as error:

        # handle error
        await system_tools.parse_error_message_with_redis(
            client_redis,
            error,
        )


async def abnormal_trading_notices_in_message_channel(
    pipe: object,
    abnormal_trading_notices_channel: str,
    pub_message: list,
    result: dict,
) -> None:

    result["params"].update({"channel": abnormal_trading_notices_channel})
    result["params"].update(pub_message)

    await redis_client.publishing_result(
        pipe,
        result,
    )
